import numpy as np
#import networkx as nx

from f90pagerank import f90pagerank as fpr

def GoogleMatrix(digraph, beta):
    nodedict = {node: idx for idx, node in zip(range(len(digraph)), digraph.nodes())}
    A = np.matrix((1 - beta) / float(len(digraph)) * np.ones(shape=(len(digraph), len(digraph))))
    for node1, node2 in digraph.edges():
        A[nodedict[node2], nodedict[node1]] += beta / float(len(list(digraph.successors(node1))))
    return A, nodedict


def L1norm(r1, r2):
    return np.sum(abs(r1 - r2))


def CalculatePageRank(digraph, beta, eps=1e-4, maxstep=1000, fortran=False):
    A, nodes = GoogleMatrix(digraph, beta)

    if fortran:
        r = fpr.compute_pagerank(A, eps, maxstep)
        nodepr = {node: r[nodes[node]] for node in nodes}
    else:
        r = np.transpose(np.matrix(np.repeat(1 / float(len(digraph)), len(digraph))))
        converged = False
        stepid = 0
        while not converged and stepid < maxstep:
            newr = A * r
            converged = (L1norm(newr, r) < eps)
            r = newr
            stepid += 1
        nodepr = {node: r[nodes[node],0] for node in nodes}
    return nodepr





'''
Get the resistance distance matrix of a simple undirected network.
See: http://en.wikipedia.org/wiki/Resistance_distance
'''

import numpy as np
from scipy.sparse import dok_matrix

default_nodes = ['Stephen', 'Sinnie', 'Elaine']
default_edges = [('Stephen', 'Sinnie'),
                 ('Elaine', 'Sinnie'),
                 ('Elaine', 'Stephen')]

class GraphResistanceDistance:
    def __init__(self, nodes=default_nodes, edges=default_edges):
        self.initializeClass(nodes, edges)
        self.Omega = self.computeResistanceDistance()
        
    def getResistance(self, node1, node2):
        if self.nodesIdx.has_key(node1) and self.nodesIdx.has_key(node2):
            idx0 = self.nodesIdx[node1]
            idx1 = self.nodesIdx[node2]
            return self.Omega[idx0, idx1]
        else:
            unknown_keys = [node for node in [node1, node2] if not self.nodesIdx.has_key(node)]
            raise Exception('Unknown key(s): '+' '.join(unknown_keys))
    
    def initializeClass(self, nodes, edges):
        self.nodes = nodes
        # all edges are unique
        self.edges = list(set([tuple(sorted(edge)) for edge in edges]))
        self.nodesIdx = {self.nodes[idx]: idx for idx in range(len(self.nodes))}

    def calculateDegreeMatrix(self):
        Dmatrix = dok_matrix((len(self.nodes), len(self.nodes)), dtype=np.float)
        for edge in self.edges:
            for node in edge:
                idx = self.nodesIdx[node]
                Dmatrix[idx, idx] += 1
        return Dmatrix
        
    def calculateAdjacencyMatrix(self):
        Amatrix = dok_matrix((len(self.nodes), len(self.nodes)), dtype=np.float)
        for edge in self.edges:
            idx0 = self.nodesIdx[edge[0]]
            idx1 = self.nodesIdx[edge[1]]
            Amatrix[idx0, idx1] = 1
            Amatrix[idx1, idx0] = 1
        return Amatrix
        
    def computeResistanceDistance(self):
        Dmatrix = self.calculateDegreeMatrix()
        Amatrix = self.calculateAdjacencyMatrix()
        Lmatrix = Dmatrix.toarray() - Amatrix.toarray()
        Lambda = np.linalg.pinv(Lmatrix)
        Omega = dok_matrix((len(self.nodes), len(self.nodes)), dtype=np.float)
        for i in range(len(self.nodes)):
            for j in range(len(self.nodes)):
                Omega[i, j] = Lambda[i, i] + Lambda[j, j] - 2 * Lambda[i, j]
        return Omega




import networkx as nx


default_nodes = ['Stephen', 'Sinnie', 'Elaine']
default_edges = [('Stephen', 'Sinnie', 0.2),
                 ('Sinnie', 'Stephen', 0.2),
                 ('Sinnie', 'Elaine', 0.3),
                 ('Elaine', 'Sinnie', 0.2),
                 ('Stephen', 'Elaine', 1.1),
                 ('Elaine', 'Stephen', 1.2)]


class SocialNetworkSimVoltage:
    def __init__(self, nodes=default_nodes, edges=default_edges):
        self.initializeClass(nodes, edges)

    def initializeClass(self, nodes, edges):
        self.constructSocialNetwork(nodes, edges)
        self.errTol = 1e-4
        self.maxSteps = 10000

    def constructSocialNetwork(self, nodes, edges):
        self.wordNet = nx.DiGraph()
        self.wordNet.add_nodes_from(nodes)
        self.wordNet.add_weighted_edges_from(edges)
        
    def checkPersonIrrelevant(self, person, person1, person2):
        path1 = nx.algorithms.shortest_path(self.wordNet,
                                            source = person1, target = person,
                                            weight='weight')
        path2 = nx.algorithms.shortest_path(self.wordNet,
                                            source = person, target = person2,
                                            weight='weight')
        intersection_paths = list(set(path1) & set(path2))
        return (len(intersection_paths) != 1)

    def initloop(self, person1, person2):
        volDict = {}
        for node in self.wordNet:
            if node == person1:
                volDict[node] = 1.0
                continue
            elif node == person2:
                volDict[node] = 0.0
                continue
            elif self.checkPersonIrrelevant(node, person1, person2):
                volDict[node] = 10.0
                continue
            distFrom1 = float(nx.shortest_path_length(self.wordNet, person1, node, weight='weight'))
            distFrom2 = float(nx.shortest_path_length(self.wordNet, node, person2, weight='weight'))
            volDict[node] = distFrom2 / (distFrom1 + distFrom2)
        return volDict

    def compute_incurrent(self, node, volDict):
        in_current = 0
        for pred in self.wordNet.predecessors(node):
            if (volDict[pred] > volDict[node]) and (volDict[pred] >= 0.0) and (volDict[pred] <= 1.0):
                potDiff = volDict[pred] - volDict[node]
                resEdge = self.wordNet[pred][node]['weight']
                in_current += potDiff / resEdge
        return in_current

    def compute_outcurrent(self, node, volDict):
        out_current = 0
        for succ in self.wordNet.successors(node):
            if (volDict[node] > volDict[succ]) and (volDict[succ] >= 0.0) and (volDict[succ] <= 1.0):
                potDiff = volDict[node] - volDict[succ]
                resEdge = self.wordNet[node][succ]['weight']
                out_current += potDiff / resEdge
        return out_current

    def average_VR(self, node, volDict):
        sumVOverR = 0.0
        numRecR = 0.0
        for pred in self.wordNet.predecessors(node):
            if (volDict[pred] > volDict[node]) and (volDict[pred] >= 0.0) and (volDict[pred] <= 1.0):
                resEdge = self.wordNet[pred][node]['weight']
                sumVOverR += volDict[pred] / resEdge
                numRecR += 1. / resEdge
        for succ in self.wordNet.successors(node):
            if (volDict[node] > volDict[succ]) and (volDict[succ] >= 0.0) and (volDict[succ] <= 1.0):
                resEdge = self.wordNet[node][succ]['weight']
                sumVOverR += volDict[succ] / resEdge
                numRecR += 1. / resEdge
        return sumVOverR, numRecR

    def getResistance(self, person1, person2, printVol = False):
        if person1 == person2:
            return 0.0
        try:
            distTwoWords = nx.shortest_path_length(self.wordNet, person1, person2, weight='weight')
        except nx.exception.NetworkXNoPath:
            return float('inf')

        # initialization
        volDict = self.initloop(person1, person2)
        if printVol:
            print(volDict)
        tempVolDict = {node: volDict[node] for node in self.wordNet}

        # iteration: computing the potential of each node
        converged = False
        step = 0
        while (not converged) and step < self.maxSteps:
            tempConverged = True
            for node in self.wordNet:
                if node == person1:
                    tempVolDict[node] = 1.0
                    continue
                elif node == person2:
                    tempVolDict[node] = 0.0
                    continue
                elif (volDict[node] < 0.0) or (volDict[node] > 1.0):
                    tempVolDict[node] = 10.0
                    continue
                in_current = self.compute_incurrent(node, volDict)
                out_current = self.compute_outcurrent(node, volDict)
                if abs(in_current - out_current) > self.errTol:
                    sumVOverR, numRecR = self.average_VR(node, volDict)
                    tempVolDict[node] = 0.0 if numRecR==0 else sumVOverR / numRecR
                    tempConverged = False
                else:
                    tempConverged = tempConverged and True
            converged = tempConverged
            # value update
            for node in self.wordNet:
                volDict[node] = tempVolDict[node]
            step += 1
            if printVol:
                print(volDict)

        # calculating the resistance
        startCurrent = sum([(1.0-volDict[rootsucc])/self.wordNet[person1][rootsucc]['weight']
                            for rootsucc in self.wordNet.successors(person1) if volDict[rootsucc]<=1.0])
        return (1.0 / startCurrent)
                                
    def drawNetwork(self):
        nx.draw(self.wordNet)
        

        





from functools import partial

from scipy.spatial import Delaunay, distance

from .abssimcomplex import SimplicialComplex
from . import facesiter, get_allpoints


def calculate_distmatrix(points, labels, distfcn):
    return {(labels[i], labels[j]): distfcn(points[i], points[j])
            for i in range(len(labels)) for j in range(len(labels))}

def contain_detachededges(simplex, distdict, epsilon):
    if len(simplex)==2:
        return (distdict[simplex[0], simplex[1]] > 2*epsilon)
    else:
        return (True in map(partial(contain_detachededges, distdict=distdict, epsilon=epsilon), facesiter(simplex)))

class AlphaComplex(SimplicialComplex):
    def __init__(self, points, epsilon, labels=None, distfcn=distance.euclidean):
        self.pts = points
        self.labels = range(len(self.pts)) if labels==None or len(labels)!=len(self.pts) else labels
        self.epsilon = epsilon
        self.distfcn = distfcn
        self.import_simplices(self.construct_simplices(self.pts, self.labels, self.epsilon, self.distfcn))

    def construct_simplices(self, points, labels, epsilon, distfcn):
        delaunay_simplices = map(tuple, Delaunay(points).simplices)
        distdict = calculate_distmatrix(points, labels, distfcn)

        simplices = []
        for simplex in delaunay_simplices:
            faces = list(facesiter(simplex))
            detached = map(partial(contain_detachededges, distdict=distdict, epsilon=epsilon), faces)
            if True in detached and len(simplex)>2:
                simplices += [face for face, notkeep in zip(faces, detached) if not notkeep]
            else:
                simplices.append(simplex)
        simplices = map(lambda simplex: tuple(sorted(simplex)), simplices)
        simplices = list(set(simplices))

        allpts = get_allpoints(simplices)
        simplices += [(point,) for point in (set(labels)-allpts)]

        return simplices


# from setuptools import setup

from numpy.distutils.core import setup, Extension

def readme():
    with open('README.md') as f:
        return f.read()


setup(name='mogu',
      version="0.1.11",
      description="Collection of Simple Numerical Routines",
      long_description="Collection of simple numerical routines, independent of each other",
      classifiers=[
          "Topic :: Scientific/Engineering :: Artificial Intelligence",
          "Topic :: Scientific/Engineering :: Mathematics",
          "Programming Language :: Python :: 2.7",
          "License :: OSI Approved :: MIT License",
      ],
      keywords="mogu numerics computation",
      url="https://github.com/stephenhky/MoguNumerics",
      author="Kwan-Yuet Ho",
      author_email="stephenhky@yahoo.com.hk",
      license='MIT',
      packages=['mogu',
                'mogu.embed',
                'mogu.fit',
                'mogu.util',
                'mogu.util.derivatives',
                'mogu.netflow',
                'mogu.netflow.simvoltage',
                'mogu.netflow.pagerank',
                'mogu.assrule',
                'mogu.voterank',
                'mogu.finance',
                'mogu.finance.binomial',
                'mogu.dynprog',
                'mogu.topology'],
      package_data={'mogu': ['finance/binomial/*.f90', 'finance/binomial/*.pyf',
                             'dynprog/*.c', 'dynprog/*.i', 'dynprog/*.h',
                             'netflow/pagerank/*.f90', 'netflow/pagerank/*.pyf']},
      setup_requires=['numpy',],
      install_requires=[
          'numpy', 'scipy', 'numba', 'tensorflow', 'networkx>=2.0',
      ],
      tests_require=[
          'unittest2',
      ],
      scripts=['bin/concatenate_dict', 'bin/mogu_minerule', 'bin/price_option', 'bin/mogu_sammon'],
      ext_modules = [Extension( 'binomialtree', sources=['mogu/finance/binomial/binomialtree.f90',
                                                         'mogu/finance/binomial/binomialtree.pyf'] ),
                     Extension( 'f90pagerank', sources=['mogu/netflow/pagerank/f90pagerank.f90',
                                                        'mogu/netflow/pagerank/f90pagerank.pyf']),
                     Extension( '_dldist', sources=['mogu/dynprog/dldist_wrap.c',
                                                    'mogu/dynprog/dldist.c']),
                     ],
      # include_package_data=True,
      # test_suite="test",
      zip_safe=False)





import unittest

from mogu.finance.binomial import eurocall_price

class test_finance(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testEuroCall(self):
        self.assertAlmostEquals(eurocall_price(100, 110, 0.01, 0.01, 10, 10000),
                                1.4968152046203613,
                                delta=1e-4)

if __name__ == '__main__':
    unittest.main()



import tensorflow as tf

def sammon_embedding(Xmat, initYmat, tol=1e-8, alpha=0.3, nbsteps=1000):
    N = Xmat.shape[0]
    # d = Xmat.shape[1]

    # distance matrix for X
    X = tf.placeholder('float')
    sqX = tf.reduce_sum(X * X, 1)
    sqX = tf.reshape(sqX, [-1, 1])
    sqDX = sqX - 2 * tf.matmul(X, tf.transpose(X)) + tf.transpose(sqX)
    sqDXarray = tf.stack([sqDX[i, j] for i in range(N) for j in range(i + 1, N)])
    DXarray = tf.sqrt(sqDXarray)

    # distance matrix for Y
    Y = tf.Variable(initYmat, dtype='float')
    sqY = tf.reduce_sum(Y * Y, 1)
    sqY = tf.reshape(sqY, [-1, 1])
    sqDY = sqY - 2 * tf.matmul(Y, tf.transpose(Y)) + tf.transpose(sqY)
    sqDYarray = tf.stack([sqDY[i, j] for i in range(N) for j in range(i + 1, N)])
    DYarray = tf.sqrt(sqDYarray)

    # cost function
    Z = tf.reduce_sum(DXarray) * 0.5
    numerator = tf.reduce_sum(tf.divide(tf.square(DXarray - DYarray), DXarray)) * 0.5
    cost = tf.divide(numerator, Z)

    # optimizer (note: do not use Adam, which can lead to overflow)
    train = tf.train.AdagradOptimizer(alpha).minimize(cost)
    init = tf.global_variables_initializer()

    # Tensorflow session
    sess = tf.Session()
    sess.run(init)

    # training
    c = sess.run(cost, feed_dict={X: Xmat})
    print "initial cost = ", c
    converged = False
    i = 0
    while (not converged) and (i < nbsteps):
        sess.run(train, feed_dict={X: Xmat})
        newc = sess.run(cost, feed_dict={X: Xmat})
        print "epoch: ", i, " cost = ", newc
        converged = (abs(newc-c)<tol)
        i += 1
        c = newc

    # result retrieval
    calculated_Y = sess.run(Y, feed_dict={X: Xmat})

    # close session
    sess.close()

    return calculated_Y